

__author__ = 'Robert Meyer'

try:
    from ._version import __version__
except ImportError:
    # We're running in a tree that doesn't
    # have a _version.py, so we don't know what our version is.
    __version__ = "unknown"


from pypet.environment import Environment, MultiprocContext
from pypet.trajectory import Trajectory, load_trajectory
from pypet.storageservice import HDF5StorageService, LazyStorageService
from pypet.naturalnaming import ParameterGroup, DerivedParameterGroup, ConfigGroup,\
    ResultGroup, NNGroupNode, NNLeafNode, KnowsTrajectory, new_group
from pypet.parameter import Parameter, ArrayParameter, SparseParameter,\
    PickleParameter, Result, SparseResult, PickleResult, ObjectTable, BaseParameter, BaseResult
from pypet.pypetexceptions import DataNotInStorageError, NoSuchServiceError,\
    NotUniqueNodeError, ParameterLockedException, PresettingError, TooManyGroupsError,\
    VersionMismatchError, GitDiffError
from pypet.pypetlogging import HasLogger, rename_log_file
from pypet.utils.explore import cartesian_product, find_unique_points
from pypet.utils.hdf5compression import compact_hdf5_file
from pypet.utils.helpful_functions import progressbar
from pypet.shareddata import SharedArray, SharedCArray, SharedEArray,\
    SharedVLArray, SharedPandasFrame, SharedTable, SharedResult,\
    StorageContextManager, make_ordinary_result, make_shared_result
from pypet.slots import HasSlots


__all__ = [
    Trajectory.__name__,
    Environment.__name__,
    MultiprocContext.__name__,
    HDF5StorageService.__name__,
    LazyStorageService.__name__,
    ParameterGroup.__name__,
    DerivedParameterGroup.__name__,
    ConfigGroup.__name__,
    ResultGroup.__name__,
    NNGroupNode.__name__,
    NNLeafNode.__name__,
    'new_group',
    BaseParameter.__name__,
    Parameter.__name__,
    ArrayParameter.__name__,
    SparseParameter.__name__,
    PickleParameter.__name__,
    BaseResult.__name__,
    Result.__name__,
    SparseResult.__name__,
    PickleResult.__name__,
    ObjectTable.__name__,
    DataNotInStorageError.__name__,
    NoSuchServiceError.__name__,
    NotUniqueNodeError.__name__,
    ParameterLockedException.__name__,
    PresettingError.__name__,
    TooManyGroupsError.__name__,
    VersionMismatchError.__name__,
    GitDiffError.__name__,
    HasSlots.__name__,
    HasLogger.__name__,
    rename_log_file.__name__,
    cartesian_product.__name__,
    load_trajectory.__name__,
    compact_hdf5_file.__name__,
    KnowsTrajectory.__name__,
    StorageContextManager.__name__,
    SharedArray.__name__,
    SharedCArray.__name__,
    SharedEArray.__name__,
    SharedVLArray.__name__,
    SharedPandasFrame.__name__,
    SharedTable.__name__,
    SharedResult.__name__,
    make_ordinary_result.__name__,
    make_shared_result.__name__,
    progressbar.__name__,
    find_unique_points.__name__
]



__author__ = 'Robert Meyer'

import brian
from brian.monitor import SpikeMonitor,SpikeCounter,StateMonitor, \
    PopulationSpikeCounter, PopulationRateMonitor, StateSpikeMonitor,  \
        MultiStateMonitor, ISIHistogramMonitor, VanRossumMetric
from brian.fundamentalunits import get_unit_fast
from brian import clear

from pypet.brian.parameter import BrianMonitorResult
from pypet.tests.unittests.parameter_test import ResultTest
from pypet.tests.unittests.briantests.run_a_brian_network import run_network
import pypet.utils.comparisons as comp
from pypet.tests.testutils.ioutils import run_suite, parse_args


class BrianMonitorTest(ResultTest):

    tags = 'unittest', 'brian', 'result', 'monitor'

    @classmethod
    def setUpClass(cls):
        BrianMonitorResult.monitors=run_network()
        pass

    @classmethod
    def tearDownClass(cls):
        clear(True, True)
        reload(brian)

    def setUp(self):
        self.monitors = BrianMonitorResult.monitors
        self.make_results()
        self.make_constructor()


    def check_spike_counter(self, res, monitor):
        self.assertTrue(comp.nested_equal(monitor.count, res.count))
        self.assertTrue(comp.nested_equal(monitor.nspikes, res.nspikes))
        self.assertTrue(comp.nested_equal(str(monitor.source), res.source))

    def check_van_rossum_metric(self,res, monitor):
        self.assertTrue(comp.nested_equal(float(monitor.tau), res.tau))
        self.assertTrue(comp.nested_equal('second', res.tau_unit))
        self.assertTrue(comp.nested_equal(monitor.distance, res.distance))
        self.assertTrue(comp.nested_equal(monitor.N, res.N))
        self.assertTrue(comp.nested_equal(str(monitor.source), res.source))

    def check_population_spike_counter(self, res, monitor):
        self.assertTrue(comp.nested_equal(monitor.delay, res.delay))
        self.assertTrue(comp.nested_equal(monitor.nspikes, res.nspikes))
        self.assertTrue(comp.nested_equal(str(monitor.source), res.source))

    def check_state_spike_monitor(self, res, monitor):
        self.assertTrue(comp.nested_equal(monitor.delay, res.delay))
        self.assertTrue(comp.nested_equal(monitor.nspikes, res.nspikes))
        self.assertTrue(comp.nested_equal(str(monitor.source), res.source))
        self.assertTrue(comp.nested_equal(monitor._varnames, res.varnames))

        self.assertEqual('second', res.spiketimes_unit)

        if res.v_storage_mode == BrianMonitorResult.TABLE_MODE:
            spike_frame = res.spikes
            spiked_list=sorted(list(set(spike_frame['neuron'].to_dict().values())))
            self.assertEqual(spiked_list, res.neurons_with_spikes)
            for idx,val_tuple in enumerate(monitor.spikes):
                neuron = val_tuple[0]
                time = val_tuple[1]
                vals = val_tuple[2:]

                self.assertEqual(neuron, spike_frame['neuron'][idx])

                self.assertEqual(float(time), spike_frame['spiketimes'][idx])

                for idx_var, varname in enumerate(res.varnames):
                    val = vals[idx_var]
                    self.assertEqual(float(val),spike_frame[varname][idx])



        elif res.v_storage_mode == BrianMonitorResult.ARRAY_MODE:

            self.assertTrue('%0' in res.format_string and 'd' in res.format_string)

            spiked_set=set()
            for item_name in res:

                if item_name.startswith('spiketimes') and not item_name.endswith('unit'):
                    neuron_id = int(item_name.split('_')[-1])
                    spiked_set.add(neuron_id)

                    times = monitor.times(neuron_id)
                    self.assertTrue(comp.nested_equal(times,res[item_name]))

                for varname in res.varnames:
                    if item_name.startswith(varname) and not item_name.endswith('unit'):
                        neuron_id =int(item_name.split('_')[-1])
                        values = monitor.values(varname,neuron_id)


                        # Remove units:
                        self.assertTrue(comp.nested_equal(values,res[item_name]))

            spiked_list = sorted(list(spiked_set))
            self.assertEqual(spiked_list, res.neurons_with_spikes)
        else:
            raise RuntimeError('You shall not pass!')


        # Check Units
        for idx,varname in enumerate(monitor._varnames):
            unit = repr(get_unit_fast(monitor.spikes[0][idx+2]))
            self.assertTrue(unit,res[varname+'_unit'])


    def check_population_rate_monitor(self, res, monitor):
        self.assertEqual(str(monitor.source),res.source)
        self.assertTrue(comp.nested_equal(monitor._bin,res.bin))
        self.assertTrue(comp.nested_equal('second',res.times_unit))
        self.assertTrue(comp.nested_equal('Hz',res.rate_unit))
        self.assertTrue(comp.nested_equal(monitor.rate,res.rate))
        self.assertTrue(comp.nested_equal(monitor.times,res.times))
        self.assertTrue(comp.nested_equal(monitor.delay,res.delay))

    def check_isi_hist_monitor(self, res, monitor):
        self.assertEqual(str(monitor.source),res.source)
        self.assertTrue(comp.nested_equal(monitor.count,res.count))
        self.assertTrue(comp.nested_equal(monitor.bins,res.bins))
        self.assertTrue(comp.nested_equal(monitor.delay,res.delay))
        self.assertEqual(monitor.nspikes, res.nspikes)


    def check_spike_monitor(self, res, monitor):
        self.assertTrue(comp.nested_equal(monitor.delay, res.delay))
        self.assertTrue(comp.nested_equal(monitor.nspikes, res.nspikes))
        self.assertTrue(comp.nested_equal(str(monitor.source), res.source))


        self.assertEqual('second', res.spiketimes_unit)

        if res.v_storage_mode == BrianMonitorResult.TABLE_MODE:
            spike_frame = res.spikes
            spiked_list=sorted(list(set(spike_frame['neuron'].to_dict().values())))
            self.assertEqual(spiked_list, res.neurons_with_spikes)
            for idx,val_tuple in enumerate(monitor.spikes):
                neuron = val_tuple[0]
                time = val_tuple[1]
                vals = val_tuple[2:]

                self.assertEqual(neuron, spike_frame['neuron'][idx])

                self.assertEqual(float(time), spike_frame['spiketimes'][idx])


        elif res.v_storage_mode == BrianMonitorResult.ARRAY_MODE:

            self.assertTrue('%0' in res.format_string and 'd' in res.format_string)

            spiked_set=set()
            for item_name in res:

                if item_name.startswith('spiketimes') and not item_name.endswith('unit'):
                    neuron_id = int(item_name.split('_')[-1])
                    spiked_set.add(neuron_id)

                    times = monitor.spiketimes[neuron_id]
                    self.assertTrue(comp.nested_equal(times,res[item_name]))

            spiked_list = sorted(list(spiked_set))
            self.assertEqual(spiked_list, res.neurons_with_spikes)
        else:
            raise RuntimeError('You shall not pass!')

    def check_multi_state_monitor(self, res, monitor):
        self.assertEqual(monitor.vars, res.vars)

        if len(monitor.times)>0:
            self.assertEqual('second', res.times_unit)
            self.assertTrue(comp.nested_equal(monitor.times, res.times))




        for idx, varname in enumerate(monitor.vars):
            mon = monitor.monitors[varname]
            self.assertTrue(comp.nested_equal(mon.record, res.record))

            self.assertTrue(comp.nested_equal(mon.when, res.when))

            self.assertEqual(mon.timestep, res.timestep)
            self.assertTrue(comp.nested_equal(str(mon.P), res.source))

            if hasattr(res, varname+'_mean'):
                self.assertTrue(comp.nested_equal(mon.mean, res.f_get(varname+'_mean')))
                self.assertTrue(comp.nested_equal(mon.var, res.f_get(varname+'_var')))
            if len(monitor.times)>0:
                self.assertTrue(comp.nested_equal(mon.values, res.f_get(varname+'_values')))

            self.assertTrue(comp.nested_equal(repr(mon.unit), res.f_get(varname+'_unit')))


    def check_state_monitor(self, res, monitor):

        self.assertEqual('second', res.times_unit)
        self.assertEqual(monitor.varname, res.varname)
        self.assertEqual(repr(monitor.unit), res.unit)
        self.assertTrue(comp.nested_equal(monitor.record, res.record))

        self.assertTrue(comp.nested_equal(monitor.when, res.when))

        self.assertEqual(monitor.timestep, res.timestep)
        self.assertTrue(comp.nested_equal(str(monitor.P), res.source))

        if hasattr(res, 'mean'):
            self.assertTrue(comp.nested_equal(monitor.mean, res.mean))
            self.assertTrue(comp.nested_equal(monitor.var, res.var))

        if len(monitor.times) > 0:
            self.assertTrue(comp.nested_equal(monitor.times, res.times))
            self.assertTrue(comp.nested_equal(monitor.values, res.values))


    def test_failing_adding_another_monitor_or_changing_the_mode(self):

        for res in self.results.values():
            with self.assertRaises(TypeError):
                res.f_set(monitor=self.monitors.values()[0])

        for res in self.results.values():
            with self.assertRaises(TypeError):
                res.v_storage_mode=BrianMonitorResult.ARRAY_MODE

    def test_the_insertion_made_implicetly_in_setUp(self):

        for key,monitor in self.monitors.items():
            res = self.results[key]
            self.assertEqual(res.testtestextra,42)
            self.assertEqual(res.v_monitor_type,monitor.__class__.__name__)

            if isinstance(monitor, SpikeCounter):
                self.check_spike_counter(res, monitor)

            elif isinstance(monitor, VanRossumMetric):
                self.check_van_rossum_metric(res, monitor)

            elif isinstance(monitor, PopulationSpikeCounter):
                self.check_population_spike_counter(res, monitor)

            elif isinstance(monitor, StateSpikeMonitor):
                self.check_state_spike_monitor(res, monitor)

            elif  isinstance(monitor, PopulationRateMonitor):
                self.check_population_rate_monitor(res, monitor)


            elif isinstance(monitor, ISIHistogramMonitor):
                self.check_isi_hist_monitor(res, monitor)

            elif isinstance(monitor,SpikeMonitor):
                self.check_spike_monitor(res, monitor)

            elif isinstance(monitor, MultiStateMonitor):
                self.check_multi_state_monitor(res, monitor)

            elif isinstance(monitor,StateMonitor):
                self.check_state_monitor(res, monitor)


            else:
                raise ValueError('Monitor Type %s is not supported (yet)' % str(type(monitor)))

    def test_dir(self):
        res = self.results['SpikeMonitor']
        self.assertTrue('spikes' in res)
        self.assertTrue('spikes' in dir(res))

    def test_set_item_via_number(self):
        res = self.results['SpikeMonitor']
        res[0] = 'Hi'
        res[777] = 777

        self.assertTrue(getattr(res, res.v_name) == 'Hi')
        self.assertTrue(res.f_get(0) == 'Hi')
        self.assertTrue(getattr(res, res.v_name + '_777') == 777)
        self.assertTrue(res[777] == 777)
        self.assertTrue(res.f_get(777) == 777)

        self.assertTrue(0 in res)
        self.assertTrue(777 in res)
        self.assertTrue(99999999 not in res)

        del res[0]
        self.assertTrue(0 not in res)

        del res[777]
        self.assertTrue(777 not in res)

    def make_results(self):
        self.results={}
        for key,monitor in self.monitors.items():
            if key.endswith('Ar'):
                self.results[key]=BrianMonitorResult(key,monitor,
                                                     storage_mode=BrianMonitorResult.ARRAY_MODE)
            else:
                self.results[key]=BrianMonitorResult(key,monitor)

            self.results[key].f_set(testtestextra=42)

if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)


"""Module containing results and parameters that can be used to store `BRIAN data`_.

Parameters handling BRIAN data are instantiated by the
:class:`~pypet.brian.parameter.BrianParameter` class for any BRIAN Quantity.

The :class:`~pypet.brian.parameter.BrianResult` can store BRIAN Quantities
and the :class:`~pypet.brian.parameter.BrianMonitorResult` extracts data from
BRIAN Monitors.

All these can be combined with the experimental framework in `pypet.brian.network` to allow
fast setup of large scale BRIAN experiments.

.. _`BRIAN data`: http://briansimulator.org/

"""

___author__ = 'Robert Meyer'


from pypet.parameter import Parameter, Result, ObjectTable
# Unfortunately I have to do this for read the docs to be able to mock brian
from pypet.utils.decorators import copydoc
import pypet.pypetexceptions as pex

try:
    from brian.units import *
    from brian.stdunits import *
except TypeError:
    pass

from brian.fundamentalunits import Quantity, get_unit_fast
from brian.monitor import SpikeMonitor, SpikeCounter, StateMonitor, \
    PopulationSpikeCounter, PopulationRateMonitor, StateSpikeMonitor,  \
    MultiStateMonitor, ISIHistogramMonitor, VanRossumMetric, Monitor

import numpy as np
import pandas as pd


class BrianParameter(Parameter):
    """A Parameter class that supports BRIAN Quantities.

    Note that only scalar BRIAN quantities are supported, lists, tuples or dictionaries
    of BRIAN quantities cannot be handled.

    There are two storage modes, that can be either passed to constructor or changed
    via `v_storage_mode`:

    * :const:`~pypet.brian.parameter.BrianParameter.FLOAT_MODE`: ('FLOAT')

        The value is stored as a float and the unit as a sting.

        i.e. `12 mV` is stored as `12.0` and `'1.0 * mV'`

    * :const:`~pypet.brian.parameter.BrianParameter.STRING_MODE`: ('STRING')

        The value and unit are stored combined together as a string.

        i.e. `12 mV` is stored as `'12.0 * mV'`


    Supports data for the standard :class:`~pypet.parameter.Parameter`, too.

    """

    IDENTIFIER = '__brn__'
    ''' Identification string stored into column title of hdf5 table'''

    FLOAT_MODE = 'FLOAT'
    '''Float storage mode'''
    STRING_MODE = 'STRING'
    '''String storage mode'''

    __slots__ = ('_storage_mode',)

    def __init__(self, full_name, data=None, comment='', storage_mode=FLOAT_MODE):
        super(BrianParameter, self).__init__(full_name, data, comment)

        self._storage_mode = None
        self.v_storage_mode = storage_mode

    @property
    def v_storage_mode(self):
        """
        There are two storage modes:


        * :const:`~pypet.brian.parameter.BrianParameter.FLOAT_MODE`: ('FLOAT')

            The value is stored as a float and the unit as a sting.

            i.e. `12 mV` is stored as `12.0` and `'1.0 * mV'`

        * :const:`~pypet.brian.parameter.BrianParameter.STRING_MODE`: ('STRING')

            The value and unit are stored combined together as a string.

            i.e. `12 mV` is stored as `'12.0 * mV'`

        """
        return self._storage_mode

    @v_storage_mode.setter
    def v_storage_mode(self, storage_mode):
        assert (storage_mode == BrianParameter.STRING_MODE or
                storage_mode == BrianParameter.FLOAT_MODE)
        self._storage_mode = storage_mode


    def f_supports(self, data):
        """ Simply checks if data is supported """
        if isinstance(data, Quantity):
            return True
        if super(BrianParameter, self).f_supports(data):
            return True
        return False

    def _values_of_same_type(self, val1, val2):

        if isinstance(val1, Quantity):
            try:
                if not val1.has_same_dimensions(val2):
                    return False
            except AttributeError:
                return False
        elif isinstance(val2, Quantity):
            try:
                if not val2.has_same_dimensions(val1):
                    return False
            except AttributeError:
                return False

        elif not super(BrianParameter, self)._values_of_same_type(val1, val2):
            return False


        return True

    def _store(self):

        if isinstance(self._data, Quantity):
            store_dict = {}

            if self._storage_mode == BrianParameter.STRING_MODE:

                valstr = self._data.in_best_unit(python_code=True)
                store_dict['data' + BrianParameter.IDENTIFIER] = ObjectTable(
                    data={'data': [valstr]})


                if self.f_has_range():
                    valstr_list = []
                    for val in self._explored_range:
                        valstr = val.in_best_unit(python_code=True)
                        valstr_list.append(valstr)


                    store_dict['explored_data' + BrianParameter.IDENTIFIER] = \
                        ObjectTable(data={'data': valstr_list})

            elif self._storage_mode == BrianParameter.FLOAT_MODE:
                unitstr = repr(get_unit_fast(self._data))
                value = float(self._data)
                store_dict['data' + BrianParameter.IDENTIFIER] = ObjectTable(
                    data={'value': [value], 'unit': [unitstr]})

                if self.f_has_range():
                    value_list = []
                    for val in self._explored_range:
                        value = float(val)
                        value_list.append(value)


                    store_dict['explored_data' + BrianParameter.IDENTIFIER] = \
                        ObjectTable(data={'value': value_list})

            else:
                raise RuntimeError('You shall not pass!')

            self._locked = True

            return store_dict
        else:
            return super(BrianParameter, self)._store()


    def _load(self, load_dict):
        if self.v_locked:
            raise pex.ParameterLockedException('Parameter `%s` is locked!' % self.v_full_name)

        try:
            data_table = load_dict['data' + BrianParameter.IDENTIFIER]

            if 'unit' in data_table:
                self._storage_mode = BrianParameter.FLOAT_MODE
            else:
                self._storage_mode = BrianParameter.STRING_MODE

            if self._storage_mode == BrianParameter.STRING_MODE:
                valstr = data_table['data'][0]

                self._data = eval(valstr)


                if 'explored_data' + BrianParameter.IDENTIFIER in load_dict:
                    explore_table = load_dict['explored_data' + BrianParameter.IDENTIFIER]

                    valstr_col = explore_table['data']
                    explore_list = []
                    for valstr in valstr_col:
                        brian_quantity = eval(valstr)
                        explore_list.append(brian_quantity)

                    self._explored_range = tuple(explore_list)
                    self._explored = True

            elif self._storage_mode == BrianParameter.FLOAT_MODE:

                # Recreate the brain units from the vale as float and unit as string:
                unit = eval(data_table['unit'][0])
                value = data_table['value'][0]
                self._data = value * unit

                if 'explored_data' + BrianParameter.IDENTIFIER in load_dict:
                    explore_table = load_dict['explored_data' + BrianParameter.IDENTIFIER]

                    value_col = explore_table['value']
                    explore_list = []
                    for value in value_col:
                        brian_quantity = value * unit
                        explore_list.append(brian_quantity)

                    self._explored_range = tuple(explore_list)
                    self._explored = True


        except KeyError:
            super(BrianParameter, self)._load(load_dict)

        self._default = self._data
        self._locked = True


class BrianDurationParameter(BrianParameter):
    """Special BRIAN parameter to specify orders and durations of subruns.

    The :class:`~pypet.brian.network.NetworkRunner` extracts the individual subruns
    for a given network from such duration parameters.
    The order of execution is defined by the property `v_order`.
    The exact values do not matter only the rank ordering.

    A Duration Parameter should be in time units (ms or s, for instance).

    DEPRECATED: Please use a normal :class:`~pypet.brian.BrianParameter` instead and
    add the property `order` to it's :class:`~pypet.annotations.Annotations`.
    No longer use:

        >>> subrun = BrianDurationParameter('mysubrun', 10 * s, order=42)

    But use:

        >>> subrun = BrianParameter('mysubrun', 10 * s)
        >>> subrun.v_annotations.order=42

    """
    def __init__(self, full_name, data=None, order=0, comment='',
                 storage_mode=BrianParameter.FLOAT_MODE):
        super(BrianDurationParameter, self).__init__(full_name, data, comment, storage_mode)
        self.v_annotations.order = order

    @property
    def v_order(self):
        """The order in which the subrun with a particular duration will be run
        by the network runner"""
        return self.v_annotations.order

    @v_order.setter
    def v_order(self, order):
        self.v_annotations.order = order

    def _load(self, load_dict):
        if 'order' in load_dict:
            self.v_annotations.order = load_dict['order']
        super(BrianDurationParameter, self)._load(load_dict)


class BrianResult(Result):
    """ A result class that can handle BRIAN quantities.

    Note that only scalar BRIAN quantities are supported, lists, tuples or dictionaries
    of BRIAN quantities cannot be handled.

    Supports also all data supported by the standard :class:`~pypet.parameter.Result`.

    Storage mode works as for :class:`~pypet.brian.parameter.BrianParameter`.

    """

    IDENTIFIER = BrianParameter.IDENTIFIER
    ''' Identifier String to label brian data '''

    FLOAT_MODE = 'FLOAT'
    '''Float storage mode'''

    STRING_MODE = 'STRING'
    '''String storage mode'''

    __slots__ = ('_storage_mode',)

    def __init__(self, full_name, *args, **kwargs):
        self._storage_mode = None
        storage_mode = kwargs.pop('storage_mode', BrianResult.FLOAT_MODE)
        self.v_storage_mode = storage_mode

        super(BrianResult, self).__init__(full_name, *args, **kwargs)

    @property
    def v_storage_mode(self):
        """
        There are two storage modes:


        * :const:`~pypet.brian.parameter.BrianResult.FLOAT_MODE`: ('FLOAT')

            The value is stored as a float and the unit as a sting,

            i.e. `12 mV` is stored as `12.0` and `'1.0 * mV'`

        * :const:`~pypet.brian.parameter.BrianResult.STRING_MODE`: ('STRING')

            The value and unit are stored combined together as a string,

            i.e. `12 mV` is stored as `'12.0 * mV'`

        """
        return self._storage_mode

    @v_storage_mode.setter
    def v_storage_mode(self, storage_mode):
        assert (storage_mode == BrianResult.STRING_MODE or storage_mode == BrianResult.FLOAT_MODE)
        self._storage_mode = storage_mode

    @copydoc(Result.f_set_single)
    def f_set_single(self, name, item):
        if BrianResult.IDENTIFIER in name:
            raise AttributeError('Your result name contains the identifier for brian data,'
                                 ' please do not use %s in your result names.' %
                                 BrianResult.IDENTIFIER)
        elif name == 'storage_mode':
            self.v_storage_mode = item
        else:
            super(BrianResult, self).f_set_single(name, item)

    def _supports(self, data):
        if isinstance(data, Quantity):
            return True
        else:
            return super(BrianResult, self)._supports(data)


    def _store(self):
        store_dict = {}
        for key in self._data:
            val = self._data[key]
            if isinstance(val, Quantity):

                if self._storage_mode == BrianResult.STRING_MODE:

                    valstr = val.in_best_unit(python_code=True)
                    store_dict[key + BrianResult.IDENTIFIER] = ObjectTable(
                        data={'data': [valstr]})

                elif self._storage_mode == BrianResult.FLOAT_MODE:
                    unitstr = repr(get_unit_fast(val))
                    value = float(val)
                    store_dict[key + BrianResult.IDENTIFIER] = ObjectTable(
                        data={'value': [value], 'unit': [unitstr]})

                else:
                    raise RuntimeError('You shall not pass!')

            else:
                store_dict[key] = val

        return store_dict


    def _load(self, load_dict):

        for key in load_dict:
            if BrianResult.IDENTIFIER in key:
                data_table = load_dict[key]

                if 'unit' in data_table:
                    self._storage_mode = BrianResult.FLOAT_MODE
                else:
                    self._storage_mode = BrianResult.STRING_MODE

                new_key = key.split(BrianResult.IDENTIFIER)[0]

                if self._storage_mode == BrianResult.STRING_MODE:
                    valstr = data_table['data'][0]

                    self._data[new_key] = eval(valstr)

                elif self._storage_mode == BrianResult.FLOAT_MODE:

                    # Recreate the brain units from the vale as float and unit as string:
                    unit = eval(data_table['unit'][0])
                    value = data_table['value'][0]
                    self._data[new_key] = value * unit
            else:
                self._data[key] = load_dict[key]


class BrianMonitorResult(Result):
    """ A Result class that supports brian monitors.

    Subclasses :class:`~pypet.parameter.Result`, NOT :class:`~pypet.brian.parameter.BrianResult`.
    The storage mode here works slightly different than in
    :class:`~pypet.brian.parameter.BrianResult` and :class:`~pypet.brian.parameter.BrianParameter`,
    see below.

    Monitor attributes are extracted and added as results with the attribute names.
    Note the original monitors are NOT stored, only their attribute/property values are kept.

    Add monitor on `__init__` via `monitor=` or via `f_set(monitor=brian_monitor)`

    **IMPORTANT**: You can only use 1 result per monitor. Otherwise a 'TypeError' is thrown.


    Example:

    >>> brian_result = BrianMonitorResult('example.brian_test_test.mymonitor',
                                            monitor=SpikeMonitor(...),
                                            storage_mode='TABLE',
                                            comment='Im a SpikeMonitor Example!')
    >>> brian_result.nspikes
    1337


    There are two storage modes in case you use the SpikeMonitor and StateSpikeMonitor:

    * :const:`~pypet.brian.parameter.BrianMonitorResult.TABLE_MODE`: ('TABLE')

        Default, information is stored into a single table where
        one column contains the neuron index, another the spiketimes and
        following columns contain variable values (in case of the StateSpikeMonitor)
        This is a very compact storage form.

    * :const:`~pypet.brian.parameter.BrianParameter.ARRAY_MODE`: ('ARRAY')

        For each neuron there will be a new hdf5 array,
        i.e. if you have many neurons your result node will have many entries.
        Note that this mode does sort everything according to the neurons but
        reading and writing of data might take muuuuuch longer compared to
        the other mode.

    Following monitors are supported and the following values are extraced:

    * SpikeCounter

        * count

            Array of spike counts for each neuron

        * nspikes

            Number of recorded spikes

        * source

            Name of source recorded from as string.

    * VanRossumMetric

        * tau

            Time constant of kernel.

        * tau_unit

            'second'

        * distance

            A square symmetric matrix containing the distances

        * N

            Number of neurons.

        * source

    * PopulationSpikeCounter

        * delay

            Recording delay

        * nspikes

        * source

    * StateSpikeMonitor

        * delay

        * nspikes

        * source

        * varnames

            Names of recorded variables as tuple of strings.

        * spiketimes_unit

            'second'

        * variablename_unit

            Unit of recorded variable as a string.
            'variablename' is mapped to the name of a recorded variable. For instance,
            if you recorded the membrane potential named 'vm' you would get a field named
            'vm_unit'.

        If you use v_storage_mode = :const:`~pypet.brian.parameter.BrianMonitorResult.TABLE_MODE`

            * spikes

                pandas DataFrame containing in the columns:

                'neuron': neuron indices

                'spiketimes': times of spiking

                'variablename': values of the recorded variables

        If you use v_storage_mode = :const:`~pypet.brian.parameter.BrianMonitorResult.ARRAY_MODE`

            * spiketimes_XXX

                spiketimes of neuron 'XXX' for each neuron you recorded from. The number of digits
                used to represent and format the neuron index are chosen automatically.

            * variablename_XXX

                Value of the recorded variable at spiketimes for neuron `XXX`

            * format_string

                String used to format the neuron index, for example `'%03d'`.


    * PopulationRateMonitor

        * times

            The times of the bins.

        * times_unit

            'second'

        * rate

            An array of rates in Hz

        * rate_unit

            'Hz'

        * source

        * bin

            The duration of a bin (in second).

        * delay

    * ISIHistogramMonitor:

        * source

        * delay

        * nspikes

        * bins

            The bins array passed at initialisation of the monitor.

        * count

            An array of length `len(bins)` counting how many ISIs were in each bin.

    * SpikeMonitor

        * delay

        * nspikes

        * source


        * spiketimes_unit

            'second'

        If you use v_storage_mode = :const:`~pypet.brian.parameter.BrianMonitorResult.TABLE_MODE`

            * spikes

                pandas DataFrame containing in the columns:

                'neuron': neuron indices

                'spiketimes': times of spiking

        If you use v_storage_mode = :const:`~pypet.brian.parameter.BrianMonitorResult.ARRAY_MODE`

            * spiketimes_XXX

                spiketimes of neuron 'XXX' for each neuron you recorded from. The number of digits
                used to represent and format the neuron index are chosen automatically.

            * format_string

                String used to format the neuron index, for example `'%03d'`.

    * StateMonitor

        * source

        * record

            What to record. Can be 'True' to record from all neurons. A single integer value
            or a list of integers.

        * when

            When recordings were made, for a list of potential values see BRIAN_.

        .. _BRIAN: http://briansimulator.org/docs/reference-monitors.html

        * timestep

            Integer defining the clock timestep a recording was made.

        * times

            Array of recording times

        * times_unit

            'second'

        * mean

            Mean value of the state variable for every neuron in the group. Only extracted if
            mean values are calculated by BRIAN. Note that for newer versions of BRIAN, means
            and variances are no longer extracted if `record` is NOT set to `False`.

        * var

            Unbiased estimated of variances of state variable for each neuron. Only extracted if
            variance values are calculated by BRIAN.

        * values

            A 2D array of the values of all recorded neurons, each row is a single neuron's value

        * unit

            The unit of the values as a string

        * varname

            Name of recorded variable

    * MultiStateMonitor

        As above but instead of values and unit, the result contains
        'varname_values' and 'varname_unit',
        where 'varname' is the name of the recorded variable.

    """

    TABLE_MODE = 'TABLE'
    '''Table storage mode for SpikeMonitor and StateSpikeMonitor'''

    ARRAY_MODE = 'ARRAY'
    '''Array storage mode, not recommended if you have many neurons!'''

    #keywords=set(['data','values','spikes','times','rate','count','mean_var',])
    __slots__ = ('_storage_mode', '_monitor_type')

    def __init__(self, full_name, *args, **kwargs):

        self._storage_mode = None
        self._monitor_type = None
        storage_mode = kwargs.pop('storage_mode', BrianMonitorResult.TABLE_MODE)
        self.v_storage_mode = storage_mode

        super(BrianMonitorResult, self).__init__(full_name, *args, **kwargs)


    def _store(self):
        store_dict = super(BrianMonitorResult, self)._store()

        if self._monitor_type is not None:
            store_dict['monitor_type'] = self._monitor_type
        if self._monitor_type in ['SpikeMonitor', 'StateSpikeMonitor']:
            store_dict['storage_mode'] = self.v_storage_mode

        return store_dict

    def _load(self, load_dict):
        if 'monitor_type' in load_dict:
            self._monitor_type = load_dict.pop('monitor_type')
        if self._monitor_type in ['SpikeMonitor', 'StateSpikeMonitor']:
            self._storage_mode = load_dict.pop('storage_mode')

        super(BrianMonitorResult, self)._load(load_dict)

    @property
    def v_storage_mode(self):
        """The storage mode for SpikeMonitor and StateSpikeMonitor

        There are two storage modes:

        * :const:`~pypet.brian.parameter.BrianMonitorResult.TABLE_MODE`: ('TABLE')

            Default, information is stored into a single table where
            the first column is the neuron index,
            second column is the spike time
            following columns contain variable values (in case of the StateSpikeMonitor)
            This is a very compact storage form.

        * :const:`~pypet.brian.parameter.BrianMonitorResult.ARRAY_MODE`: ('ARRAY')

            For each neuron there will be a new hdf5 array,
            i.e. if you have many neurons your result node will have many entries.
            Note that this mode does sort everything according to the neurons but
            reading and writing of data might take muuuuuch longer compared to
            the other mode.

        """
        return self._storage_mode

    @property
    def v_monitor_type(self):
        """ The type of the stored monitor. Each MonitorResult can only manage a single Monitor.
        """
        return self._monitor_type

    @v_storage_mode.setter
    def v_storage_mode(self, storage_mode):

        if self._monitor_type is not None:
            raise TypeError('Cannot change the storage mode if you already extracted a monitor.')

        assert (storage_mode == BrianMonitorResult.ARRAY_MODE or
                storage_mode == BrianMonitorResult.TABLE_MODE)
        self._storage_mode = storage_mode

    def f_set_single(self, name, item):
        """ To add a monitor use `f_set_single('monitor', brian_monitor)`.

        Otherwise `f_set_single` works similar to :func:`~pypet.parameter.Result.f_set_single`.
        """

        if isinstance(item, (Monitor, MultiStateMonitor)):

            if self.v_stored:
                self._logger.warning('You are changing an already stored result. If '
                                     'you not explicitly overwrite the data on disk, '
                                     'this change might be lost and not propagated to disk.')

            self._extract_monitor_data(item)
        else:
            super(BrianMonitorResult, self).f_set_single(name, item)


    
    def _extract_monitor_data(self, monitor):

        if self._monitor_type is not None:
            raise TypeError('Your result `%s` already extracted data from a `%s` monitor.'
                             ' Please use a new empty result for a new monitor.')

        self._monitor_type = monitor.__class__.__name__

        ## Check for each monitor separately:
        if isinstance(monitor, SpikeCounter):
            self._extract_spike_counter(monitor)

        elif isinstance(monitor, VanRossumMetric):
            self._extract_van_rossum_metric(monitor)

        elif isinstance(monitor, PopulationSpikeCounter):
            self._extract_population_spike_counter(monitor)

        elif isinstance(monitor, StateSpikeMonitor):
            self._extract_state_spike_monitor(monitor)
        
        elif isinstance(monitor, PopulationRateMonitor):
            self._extract_population_rate_monitor(monitor)

        elif isinstance(monitor, ISIHistogramMonitor):
            self._extract_isi_hist_monitor(monitor)

        elif isinstance(monitor, SpikeMonitor):
            self._extract_spike_monitor(monitor)

        elif isinstance(monitor, MultiStateMonitor):
            self._extract_multi_state_monitor(monitor)
        
        elif isinstance(monitor, StateMonitor):
            self._extract_state_monitor(monitor)

            
        else:
            raise ValueError('Monitor Type %s is not supported (yet)' % str(type(monitor)))

    def _extract_spike_counter(self, monitor):
        self.f_set(count=monitor.count)
        self.f_set(source=str(monitor.source))
        self.f_set(delay=monitor.delay)
        self.f_set(nspikes=monitor.nspikes)


    def _extract_van_rossum_metric(self, monitor):

        self.f_set(source=str(monitor.source))
        self.f_set(N=monitor.N)
        self.f_set(tau=float(monitor.tau))
        self.f_set(tau_unit='second')
        #self.f_set(timestep = monitor.timestep)

        self.f_set(distance=monitor.distance)



    def _extract_isi_hist_monitor(self, monitor):

        self.f_set(source=str(monitor.source))
        self.f_set(count=monitor.count)
        self.f_set(bins=monitor.bins)
        self.f_set(delay=monitor.delay)
        self.f_set(nspikes=monitor.nspikes)

    @staticmethod
    def _get_format_string(monitor):
        digits = len(str(len(monitor.source)))
        format_string = '%0' + str(digits) + 'd'
        return format_string

    def _extract_state_spike_monitor(self, monitor):

        self.f_set(source=str(monitor.source))

        varnames = monitor._varnames
        if not isinstance(varnames, tuple):
            varnames = (varnames,)

        for idx, varname in enumerate(varnames):
            unit = repr(get_unit_fast(monitor.spikes[0][idx + 2]))
            self.f_set(**{varname + '_unit': unit})


        self.f_set(varnames=varnames)

        #self.f_set(record = monitor.record)
        self.f_set(delay=monitor.delay)
        self.f_set(nspikes=monitor.nspikes)
        self.f_set(spiketimes_unit='second')


        if self._storage_mode == BrianMonitorResult.TABLE_MODE:
            spike_dict = {}

            if len(monitor.spikes) > 0:
                zip_lists = list(zip(*monitor.spikes))
                time_list = zip_lists[1]

                nounit_list = [np.float64(time) for time in time_list]

                spike_dict['spiketimes'] = nounit_list
                spike_dict['neuron'] = list(zip_lists[0])

                spiked_neurons = sorted(list(set(spike_dict['neuron'])))
                if spiked_neurons:
                    self.f_set(neurons_with_spikes=spiked_neurons)

                    count = 2
                    for varname in varnames:

                        var_list = list(zip_lists[count])

                        nounit_list = [np.float64(var) for var in var_list]
                        spike_dict[varname] = nounit_list
                        count += 1

                    self.f_set(spikes=pd.DataFrame(data=spike_dict))

        elif self._storage_mode == BrianMonitorResult.ARRAY_MODE:

            format_string = self._get_format_string(monitor)
            self.f_set(format_string=format_string)

            spiked_neurons = set()

            for neuron in range(len(monitor.source)):
                spikes = monitor.times(neuron)
                if len(spikes) > 0:

                    spiked_neurons.add(neuron)

                    key = 'spiketimes_' + format_string % neuron
                    self.f_set(**{key: spikes})

            spiked_neurons = sorted(list(spiked_neurons))
            if spiked_neurons:
                self.f_set(neurons_with_spikes=spiked_neurons)

            for varname in varnames:
                for neuron in range(len(monitor.source)):
                    values = monitor.values(varname, neuron)
                    if len(values) > 0:
                        key = varname + '_' + format_string % neuron
                        self.f_set(**{key: values})
        else:
            raise RuntimeError('You shall not pass!')

    def _extract_spike_monitor(self, monitor):
        
        #assert isinstance(monitor, SpikeMonitor)


        self.f_set(source=str(monitor.source))

        self.f_set(record=monitor.record)

        self.f_set(nspikes=monitor.nspikes)

        self.f_set(spiketimes_unit='second')


        self.f_set(delay=monitor.delay)


        if self._storage_mode == BrianMonitorResult.TABLE_MODE:
            spike_dict = {}

            if len(monitor.spikes) > 0:
                zip_lists = list(zip(*monitor.spikes))
                time_list = zip_lists[1]

                nounit_list = [np.float64(time) for time in time_list]

                spike_dict['spiketimes'] = nounit_list
                spike_dict['neuron'] = list(zip_lists[0])

                spiked_neurons = sorted(list(set(spike_dict['neuron'])))
                if spiked_neurons:
                    self.f_set(neurons_with_spikes=spiked_neurons)

                    spikeframe = pd.DataFrame(data=spike_dict)
                    self.f_set(spikes=spikeframe)

        elif self._storage_mode == BrianMonitorResult.ARRAY_MODE:
            format_string = self._get_format_string(monitor)
            self.f_set(format_string=format_string)

            spiked_neurons = set()

            for neuron, spikes in monitor.spiketimes.items():
                if len(spikes) > 0:

                    spiked_neurons.add(neuron)

                    key = 'spiketimes_' + format_string % neuron
                    self.f_set(**{key: spikes})

            spiked_neurons = sorted(list(spiked_neurons))
            if spiked_neurons:
                self.f_set(neurons_with_spikes=spiked_neurons)

        else:
            raise RuntimeError('You shall not pass!')

    def _extract_population_rate_monitor(self, monitor):
        assert isinstance(monitor, PopulationRateMonitor)
        

        self.f_set(source=str(monitor.source))
        self.f_set(times_unit='second', rate_unit='Hz')
        self.f_set(times=monitor.times)
        self.f_set(rate=monitor.rate)
        self.f_set(delay=monitor.delay)
        self.f_set(bin=monitor._bin)

    def _extract_population_spike_counter(self, monitor):
        
        assert isinstance(monitor, PopulationSpikeCounter)

        self.f_set(nspikes=monitor.nspikes)
        self.f_set(source=str(monitor.source))
        self.f_set(delay=monitor.delay)




    def _extract_multi_state_monitor(self, monitors):

        self.f_set(vars=monitors.vars)


        if len(monitors.times) > 0:
            self.f_set(times=monitors.times)
            self.f_set(times_unit='second')

        ### Store recorded values ###
        for idx, varname in enumerate(monitors.vars):
            monitor = monitors.monitors[varname]
            if idx == 0:


                self.f_set(record=monitor.record)

                self.f_set(when=monitor.when)

                self.f_set(timestep=monitor.timestep)

                self.f_set(source=str(monitor.P))

            if np.mean(monitor.mean) != 0.0:
                self.f_set(**{varname + '_mean': monitor.mean})
                self.f_set(**{varname + '_var': monitor.var})
            if len(monitors.times) > 0:
                self.f_set(**{varname + '_values': monitor.values})
            self.f_set(**{varname + '_unit': repr(monitor.unit)})


    def _extract_state_monitor(self, monitor):

        self.f_set(varname=monitor.varname)
        self.f_set(unit=repr(monitor.unit))

            
        self.f_set(record=monitor.record)
        
        self.f_set(when=monitor.when)
        
        self.f_set(timestep=monitor.timestep)

        self.f_set(source=str(monitor.P))
        self.f_set(times_unit='second')

        if np.mean(monitor.mean) != 0.0:
            self.f_set(mean=monitor.mean)
            self.f_set(var=monitor.var)
        if len(monitor.times) > 0:
            self.f_set(times=monitor.times)
            self.f_set(values=monitor.values)




__author__ = 'Robert Meyer'


from pypet.tests.testutils.ioutils import run_suite, discover_tests, TEST_IMPORT_ERROR, parse_args

tests_include=set(('MultiprocNoPoolLockTest',
                   'MultiprocSortQueueTest',
                   'MultiprocLinkLockTest',
                   'CapTest',
                   'MultiprocStorageLockTest',
                   'MultiprocNoPoolQueueLoggingTest'))
big_suite_1 = discover_tests(lambda  class_name, test_name, tags: class_name in tests_include)

tests_include=set(('MultiprocNoPoolQueueTest',
                   'MultiprocSortLockTest',
                   'MultiprocLinkNoPoolLockTest',
                   'TestMPPostProc',
                   'ContinueMPPoolTest',
                   'MultiprocPoolLockLoggingTest'))
big_suite_2 = discover_tests(lambda  class_name, test_name, tags: class_name in tests_include)

tests_include=set(('MultiprocLockTest',
                   'MultiprocNoPoolSortQueueTest',
                   'MultiprocLinkNoPoolQueueTest',
                   'TestMPImmediatePostProc',
                   'MultiprocStorageNoPoolLockTest',
                   'MultiprocNoPoolLockLoggingTest'))
big_suite_3 = discover_tests(lambda  class_name, test_name, tags: class_name in tests_include)

tests_include=set(('MultiprocQueueTest',
                   'MultiprocNoPoolSortLockTest',
                   'MultiprocLinkQueueTest',
                   'ContinueMPTest',
                   'BrianFullNetworkMPTest',
                   'MultiprocPoolQueueLoggingTest'))
big_suite_4 = discover_tests(lambda  class_name, test_name, tags: class_name in tests_include)


suite_dict = {'1': big_suite_1, '2': big_suite_2, '3': big_suite_3, '4': big_suite_4}


if __name__ == '__main__':
    opt_dict = parse_args()
    suite = None
    if 'suite_no' in opt_dict:
        suite_no = opt_dict.pop('suite_no')
        suite = suite_dict[suite_no]

    if suite is None:
        pred = lambda class_name, test_name, tags: ('multiproc' in tags and
                                                    class_name != TEST_IMPORT_ERROR)
        suite = discover_tests(pred)

    run_suite(suite=suite, **opt_dict)


"""Module containing utilities for logging."""

__author__ = 'Robert Meyer'

try:
    import ConfigParser as cp
except ImportError:
    import configparser as cp
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import logging
from logging.config import fileConfig
try:
    from logging.config import dictConfig
except ImportError:
    from logutils.dictconfig import dictConfig
try:
    from logging import NullHandler
except ImportError:
    from logutils import NullHandler
import os
import math
import sys
import ast
import copy
import multiprocessing as multip
import functools

import pypet.pypetconstants as pypetconstants
import pypet.compat as compat
from pypet.utils.helpful_functions import progressbar
from pypet.slots import HasSlots


FILENAME_INDICATORS = set([pypetconstants.LOG_ENV,
                           pypetconstants.LOG_PROC,
                           pypetconstants.LOG_TRAJ,
                           pypetconstants.LOG_RUN,
                           '.log',
                           '.txt'])
"""Set of strings that mark a log file"""

LOGGING_DICT = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'file': {
            'format': '%(asctime)s %(name)s %(levelname)-8s %(message)s'
        },
        'stream': {
            'format': '%(processName)-10s %(name)s %(levelname)-8s %(message)s'
        }
    },
    'handlers': {
        'stream': {
            'class': 'logging.StreamHandler',
            'formatter': 'stream'
        },
        'file_main': {
            'class': 'logging.FileHandler',
            'formatter': 'file',
            'filename': os.path.join(pypetconstants.LOG_TRAJ,
                                     pypetconstants.LOG_ENV,
                                     'LOG.txt')
        },
        'file_error': {
            'class': 'logging.FileHandler',
            'formatter': 'file',
            'filename': os.path.join(pypetconstants.LOG_TRAJ,
                                     pypetconstants.LOG_ENV,
                                     'ERROR.txt'),
            'level': 'ERROR'
        }
    },
    'multiproc_formatters': {
        'file': {
            'format': '%(asctime)s %(name)s %(levelname)-8s %(message)s'
        },
    },
    'multiproc_handlers': {
        'file_main': {
            'class': 'logging.FileHandler',
            'formatter': 'file',
            'filename': os.path.join(pypetconstants.LOG_TRAJ,
                                     pypetconstants.LOG_ENV,
                                     '%s_%s_LOG.txt' % (pypetconstants.LOG_RUN,
                                                        pypetconstants.LOG_PROC))
        },
        'file_error': {
            'class': 'logging.FileHandler',
            'formatter': 'file',
            'filename': os.path.join(pypetconstants.LOG_TRAJ,
                                     pypetconstants.LOG_ENV,
                                     '%s_%s_ERROR.txt' % (pypetconstants.LOG_RUN,
                                                          pypetconstants.LOG_PROC)),
            'level': 'ERROR'
        }
    }
}
"""Dictionary containing the default configuration."""


def _change_logging_kwargs(kwargs):
    """ Helper function to turn the simple logging kwargs into a `log_config`."""
    log_folder = kwargs.pop('log_folder', 'logs')
    logger_names = kwargs.pop('logger_names', '')
    log_levels = kwargs.pop('log_levels', logging.INFO)

    if not isinstance(logger_names, (tuple, list)):
        logger_names = [logger_names]
    if not isinstance(log_levels, (tuple, list)):
        log_levels = [log_levels]
    if len(log_levels) == 1:
        log_levels = [log_levels[0] for _ in logger_names]

    # We don't want to manipulate the original dictionary
    dictionary = copy.deepcopy(LOGGING_DICT)

    # Add all handlers to all loggers
    for prefix in ('', 'multiproc_'):
        for handler_dict in dictionary[prefix + 'handlers'].values():
            if 'filename' in handler_dict:
                handler_dict['filename'] = os.path.join(log_folder, handler_dict['filename'])
        dictionary[prefix + 'loggers'] = {}
        logger_dict = dictionary[prefix + 'loggers']
        for idx, logger_name in enumerate(logger_names):
            logger_dict[logger_name] = {
                'level': log_levels[idx],
                'handlers': list(dictionary[prefix + 'handlers'].keys())
            }

    kwargs['log_config'] = dictionary


def simple_logging_config(func):
    """Decorator to allow a simple logging configuration.

    This encompasses giving a `log_folder`, `logger_names` as well as `log_levels`.

    """

    @functools.wraps(func)
    def new_func(self, *args, **kwargs):

        inside = [x in kwargs for x in ('log_folder', 'logger_names', 'log_levels')]
        if any(inside):
            if 'log_config' in kwargs and kwargs['log_config'] is not None:
                raise ValueError('Please set `log_config to `None` if you want to use the simple '
                                 'way of providing logging configuration.')
            _change_logging_kwargs(kwargs)

        return func(self, *args, **kwargs)

    return new_func


def try_make_dirs(filename):
    """ Tries to make directories for a given `filename`.

    Ignores any error but notifies via stderr.

    """
    try:
        dirname = os.path.dirname(filename)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
    except Exception as exc:
        sys.stderr.write('ERROR during log config file handling, could not create dirs for '
                         'filename `%s` because of: %s' % (filename, repr(exc)))


def rename_log_file(traj, filename, process_name=None):
    """ Renames a given `filename` with valid wildcard placements.

    :const:`~pypet.pypetconstants.LOG_ENV` ($env) is replaces by the name of the
    trajectory`s environment.

    :const:`~pypet.pypetconstants.LOG_TRAJ` ($traj) is replaced by the name of the
    trajectory.

    :const:`~pypet.pypetconstants.LOG_RUN` ($run) is replaced by the name of the current
    run. If the trajectory is not set to a run 'run_ALL' is used.

    :const:`~pypet.pypetconstants.LOG_SET` ($set) is replaced by the name of the current
    run set. If the trajectory is not set to a run 'run_set_ALL' is used.

    :const:`~pypet.pypetconstants.LOG_PROC` ($proc) is replaced by the name fo the
    current process.

    :param traj:  A trajectory container
    :param filename:  A filename string
    :param process_name:

        The name of the desired process. If `None` the name of the current process is
        taken determined by the multiprocessing module.

    :return: The new filename

    """
    if pypetconstants.LOG_ENV in filename:
        env_name = traj.v_environment_name
        filename = filename.replace(pypetconstants.LOG_ENV, env_name)
    if pypetconstants.LOG_TRAJ in filename:
        traj_name = traj.v_name
        filename = filename.replace(pypetconstants.LOG_TRAJ, traj_name)
    if pypetconstants.LOG_RUN in filename:
        run_name = traj.f_wildcard('$')
        filename = filename.replace(pypetconstants.LOG_RUN, run_name)
    if pypetconstants.LOG_SET in filename:
        set_name = traj.f_wildcard('$set')
        filename = filename.replace(pypetconstants.LOG_SET, set_name)
    if pypetconstants.LOG_PROC in filename:
        if process_name is None:
            process_name = multip.current_process().name
        filename = filename.replace(pypetconstants.LOG_PROC, process_name)
    return filename


def get_strings(args):
    """Returns all valid python strings inside a given argument string."""
    string_list = []
    for elem in ast.walk(ast.parse(args)):
        if isinstance(elem, ast.Str):
            string_list.append(elem.s)
    return string_list


class HasLogger(HasSlots):
    """Abstract super class that automatically adds a logger to a class.

    To add a logger to a sub-class of yours simply call ``myobj._set_logger(name)``.
    If ``name=None`` the logger is chosen as follows:

        ``self._logger = logging.getLogger(self.__class.__.__module__ + '.' + self.__class__.__name__)``

    The logger can be accessed via ``myobj._logger``.

    """

    __slots__ = ('_logger',)

    def __getstate__(self):
        """Called for pickling.

        Removes the logger to allow pickling and returns a copy of `__dict__`.

        """
        state_dict = super(HasLogger, self).__getstate__()
        if '_logger' in state_dict:
            # Pickling does not work with loggers objects, so we just keep the logger's name:
            state_dict['_logger'] = self._logger.name
        return state_dict

    def __setstate__(self, statedict):
        """Called after loading a pickle dump.

        Restores `__dict__` from `statedict` and adds a new logger.

        """
        super(HasLogger, self).__setstate__(statedict)
        if '_logger' in statedict:
            # If we re-instantiate the component the logger attribute only contains a name,
            # so we also need to re-create the logger:
            self._set_logger(statedict['_logger'])

    def _set_logger(self, name=None):
        """Adds a logger with a given `name`.

        If no name is given, name is constructed as
        `type(self).__name__`.

        """
        if name is None:
            cls = self.__class__
            name = '%s.%s' % (cls.__module__, cls.__name__)
        self._logger = logging.getLogger(name)


class LoggingManager(object):
    """ Manager taking care of all logging related issues.

    :param trajectory: Trajectory container of Mock
    :param log_config: Logging configuration

        Can be a a full name of an `ini` file. An already instantiated config parser,
        or a logging dictionary.

    :param log_stdout: If `stdout` should be logged.

    :param report_progress:

        How to report progress.

    """
    def __init__(self, trajectory=None, log_config=None, log_stdout=False,
                 report_progress=False):
        self.trajectory = trajectory
        self.log_config = log_config
        self._sc_config = None
        self._mp_config = None
        self.log_stdout = log_stdout
        self.report_progress = report_progress
        self._tools = []
        self._null_handler = NullHandler()
        self._format_string = 'PROGRESS: Finished %d/%d runs '
        # Format string for the progressbar

    def __getstate__(self):
        """ConfigParsers are not guaranteed to be picklable so we need to remove these."""
        state_dict = self.__dict__.copy()
        if isinstance(state_dict['log_config'], cp.RawConfigParser):
            # Config Parsers are not guaranteed to be picklable
            state_dict['log_config'] = True
        return state_dict

    def show_progress(self, n, total_runs):
        """Displays a progressbar"""
        if self.report_progress:
            percentage, logger_name, log_level = self.report_progress
            if logger_name == 'print':
                logger = 'print'
            else:
                logger = logging.getLogger(logger_name)

            if n == 0:
                # Compute the number of digits and avoid log10(0)
                digits = int(math.log10(total_runs + 0.1)) + 1
                self._format_string = 'PROGRESS: Finished %' + '%d' % digits + 'd/%d runs '

            fmt_string = self._format_string % (n + 1, total_runs) + '%s'
            reprint = log_level == 0
            progressbar(n, total_runs, percentage_step=percentage,
                        logger=logger, log_level=log_level,
                        fmt_string=fmt_string, reprint=reprint)

    def add_null_handler(self):
        """Adds a NullHandler to the root logger.

        This is simply added to avoid warnings that no logger has been configured.

        """
        root = logging.getLogger()
        root.addHandler(self._null_handler)

    def remove_null_handler(self):
        """Removes the NullHandler from the root logger."""
        root = logging.getLogger()
        root.removeHandler(self._null_handler)

    @staticmethod
    def tabula_rasa():
        """Removes all loggers and logging handlers. """
        erase_dict = {'disable_existing_loggers': False, 'version': 1}
        dictConfig(erase_dict)

    @staticmethod
    def _check_and_replace_parser_args(parser, section, option, rename_func, make_dirs=True):
        """ Searches for parser settings that define filenames.

        If such settings are found, they are renamed according to the wildcard
        rules. Moreover, it is also tried to create the corresponding folders.

        :param parser:  A config parser
        :param section: A config section
        :param option: The section option
        :param rename_func: A function to rename found files
        :param make_dirs: If the directories of the file should be created.

        """
        args = parser.get(section, option, raw=True)
        strings = get_strings(args)
        replace = False
        for string in strings:
            isfilename = any(x in string for x in FILENAME_INDICATORS)
            if isfilename:
                newstring = rename_func(string)
                if make_dirs:
                    try_make_dirs(newstring)
                # To work with windows path specifications we need this replacement:
                raw_string = string.replace('\\', '\\\\')
                raw_newstring = newstring.replace('\\', '\\\\')
                args = args.replace(raw_string, raw_newstring)
                replace = True
        if replace:
            parser.set(section, option, args)

    @staticmethod
    def _parser_to_string_io(parser):
        """Turns a ConfigParser into a StringIO stream."""
        memory_file = StringIO()
        parser.write(memory_file)
        memory_file.flush()
        memory_file.seek(0)
        return memory_file

    @staticmethod
    def _find_multiproc_options(parser):
        """ Searches for multiprocessing options within a ConfigParser.

        If such options are found, they are copied (without the `'multiproc_'` prefix)
        into a new parser.

        """
        sections = parser.sections()
        if not any(section.startswith('multiproc_') for section in sections):
            return None
        mp_parser = NoInterpolationParser()
        for section in sections:
            if section.startswith('multiproc_'):
                new_section = section.replace('multiproc_', '')
                mp_parser.add_section(new_section)
                options = parser.options(section)
                for option in options:
                    val = parser.get(section, option, raw=True)
                    mp_parser.set(new_section, option, val)
        return mp_parser

    @staticmethod
    def _find_multiproc_dict(dictionary):
        """ Searches for multiprocessing options in a given `dictionary`.

        If found they are copied (without the `'multiproc_'` prefix)
        into a new dictionary

        """
        if not any(key.startswith('multiproc_') for key in dictionary.keys()):
            return None
        mp_dictionary = {}
        for key in dictionary.keys():
            if key.startswith('multiproc_'):
                new_key = key.replace('multiproc_', '')
                mp_dictionary[new_key] = dictionary[key]
        mp_dictionary['version'] = dictionary['version']
        if 'disable_existing_loggers' in dictionary:
            mp_dictionary['disable_existing_loggers'] = dictionary['disable_existing_loggers']
        return mp_dictionary

    def check_log_config(self):
        """ Checks and converts all settings if necessary passed to the Manager.

        Searches for multiprocessing options as well.

        """
        if self.report_progress:
            if self.report_progress is True:
                self.report_progress = (5, 'pypet', logging.INFO)
            elif isinstance(self.report_progress, (int, float)):
                self.report_progress = (self.report_progress, 'pypet', logging.INFO)
            elif isinstance(self.report_progress, compat.base_type):
                self.report_progress = (5, self.report_progress, logging.INFO)
            elif len(self.report_progress) == 2:
                self.report_progress = (self.report_progress[0], self.report_progress[1],
                                        logging.INFO)

        if self.log_config:
            if self.log_config == pypetconstants.DEFAULT_LOGGING:
                pypet_path = os.path.abspath(os.path.dirname(__file__))
                init_path = os.path.join(pypet_path, 'logging')
                if os.sep == '\\':
                    # Use the windows setting
                    # The only differences is the path separator `\`
                    self.log_config = os.path.join(init_path, 'windows_default.ini')
                else:
                    self.log_config = os.path.join(init_path, 'default.ini')

            if isinstance(self.log_config, compat.base_type):
                if not os.path.isfile(self.log_config):
                    raise ValueError('Could not find the logger init file '
                                     '`%s`.' % self.log_config)

            if isinstance(self.log_config, compat.base_type):
                parser = NoInterpolationParser()
                parser.read(self.log_config)
            elif isinstance(self.log_config, cp.RawConfigParser):
                parser = self.log_config
            else:
                parser = None

            if parser is not None:
                self._sc_config = self._parser_to_string_io(parser)
                self._mp_config = self._find_multiproc_options(parser)
                if self._mp_config is not None:
                    self._mp_config = self._parser_to_string_io(self._mp_config)

            elif isinstance(self.log_config, dict):
                self._sc_config = self.log_config
                self._mp_config = self._find_multiproc_dict(self._sc_config)

        if self.log_stdout:
            if self.log_stdout is True:
                self.log_stdout = ('STDOUT', logging.INFO)
            if isinstance(self.log_stdout, compat.base_type):
                self.log_stdout = (self.log_stdout, logging.INFO)
            if isinstance(self.log_stdout, int):
                self.log_stdout = ('STDOUT', self.log_stdout)

    def _handle_config_parsing(self, log_config):
        """ Checks for filenames within a config file and translates them.

        Moreover, directories for the files are created as well.

        :param log_config: Config file as a stream (like StringIO)

        """
        parser = NoInterpolationParser()
        parser.readfp(log_config)

        rename_func = lambda string: rename_log_file(self.trajectory, string)

        sections = parser.sections()
        for section in sections:
            options = parser.options(section)
            for option in options:
                if option == 'args':
                    self._check_and_replace_parser_args(parser, section, option,
                                                        rename_func=rename_func)
        return parser

    def _handle_dict_config(self, log_config):
        """Recursively walks and copies the `log_config` dict and searches for filenames.

        Translates filenames and creates directories if necessary.

        """
        new_dict = dict()
        for key in log_config.keys():
            if key == 'filename':
                filename = log_config[key]
                filename = rename_log_file(self.trajectory, filename)
                new_dict[key] = filename
                try_make_dirs(filename)
            elif isinstance(log_config[key], dict):
                inner_dict = self._handle_dict_config(log_config[key])
                new_dict[key] = inner_dict
            else:
                new_dict[key] = log_config[key]
        return new_dict

    def make_logging_handlers_and_tools(self, multiproc=False):
        """Creates logging handlers and redirects stdout."""

        log_stdout = self.log_stdout
        if multiproc and hasattr(os, 'fork'):
            # If we allow forking and it is possible we already have a redirection of stdout
            log_stdout = False

        if self.log_config:
            if multiproc:
                proc_log_config = self._mp_config
            else:
                proc_log_config = self._sc_config

            if proc_log_config:
                if isinstance(proc_log_config, dict):
                    new_dict = self._handle_dict_config(proc_log_config)
                    dictConfig(new_dict)
                else:
                    parser = self._handle_config_parsing(proc_log_config)
                    memory_file = self._parser_to_string_io(parser)
                    fileConfig(memory_file, disable_existing_loggers=False)

        if log_stdout:
            #  Create a logging mock for stdout
            std_name, std_level = self.log_stdout

            stdout = StdoutToLogger(std_name, log_level=std_level)
            stdout.start()
            self._tools.append(stdout)

    def finalize(self, remove_all_handlers=True):
        """Finalizes the manager, closes and removes all handlers if desired."""
        for tool in self._tools:
            tool.finalize()
        self._tools = []
        for config in (self._sc_config, self._mp_config):
            if hasattr(config, 'close'):
                config.close()
        self._sc_config = None
        self._mp_config = None
        if remove_all_handlers:
            self.tabula_rasa()


class NoInterpolationParser(cp.ConfigParser):
    """Dummy class to solve a Python 3 bug/feature in configparser.py"""
    def __init__(self):
        try:
            # Needed for Python 3, see [http://bugs.python.org/issue21265]
            super(NoInterpolationParser, self).__init__(interpolation=None)
        except TypeError:
            # Python 2.x
            cp.ConfigParser.__init__(self)


class DisableLogger(object):
    """Context Manager that disables logging"""

    def __enter__(self):
        logging.disable(logging.CRITICAL)

    def __exit__(self, exc_type, exc_val, exc_tb):
        logging.disable(logging.NOTSET)


class StdoutToLogger(HasLogger):
    """Fake file-like stream object that redirects writes to a logger instance."""
    def __init__(self, logger_name, log_level=logging.INFO):
        self._logger_name = logger_name
        self._log_level = log_level
        self._linebuf = ''
        self._recursion = False
        self._redirection = False
        self._original_steam = None
        #self._logger = logging.getLogger(self._logger_name)
        self._set_logger(name=self._logger_name)

    def __getstate__(self):
        state_dict = super(StdoutToLogger, self).__getstate__()
        # The original stream cannot be pickled
        state_dict['_original_stream'] = None

    def start(self):
        """Starts redirection of `stdout`"""
        if sys.stdout is not self:
            self._original_steam = sys.stdout
            sys.stdout = self
            self._redirection = True
        if self._redirection:
            print('Established redirection of `stdout`.')

    def write(self, buf):
        """Writes data from buffer to logger"""
        if not self._recursion:
            self._recursion = True
            try:
                for line in buf.rstrip().splitlines():
                    self._logger.log(self._log_level, line.rstrip())
            finally:
                self._recursion = False
        else:
            # If stderr is redirected to stdout we can avoid further recursion by
            sys.__stderr__.write('ERROR: Recursion in Stream redirection!')

    def flush(self):
        """No-op to fulfil API"""
        pass

    def finalize(self):
        """Disables redirection"""
        if self._original_steam is not None and self._redirection:
            sys.stdout = self._original_steam
            print('Disabled redirection of `stdout`.')
            self._redirection = False
            self._original_steam = None

